@@ -63,6 +63,14 @@ all_opt = {
 		"required" : "0",
 		"shortdesc" : "Write debug information to given file",
 		"order" : 52 },
+	"delay" : {
+		"getopt" : "f:",
+		"longopt" : "delay",
+		"help" : "--delay <seconds>              Wait X seconds before fencing is started",
+		"required" : "0",
+		"shortdesc" : "Wait X seconds before fencing is started",
+		"default" : "0",
+		"order" : 200 },
 	"agent"   : {
 		"getopt" : "",
 		"help" : "",
@@ -369,7 +377,7 @@ all_opt = {
 		"order" : 200 }
 }
 
-common_opt = [ "retry_on" ]
+common_opt = [ "retry_on", "delay" ]
 
 class fspawn(pexpect.spawn):
 	def log_expect(self, options, pattern, timeout):
@@ -513,6 +521,11 @@ def process_input(avail_opt):
 			else:
 				longopt_list.append(all_opt[k]["longopt"])
 
+	## Compatibility layer
+	if avail_opt.count("module_name") == 1:
+		getopt_string += "n:"
+		longopt_list.append("plug=")
+	
 	##
 	## Read options from command line or standard input
 	#####
@@ -539,6 +552,8 @@ def process_input(avail_opt):
 		z = dict(opt)
 		if z.has_key("-T") == 1:
 			z["-o"] = "status"
+		if z.has_key("-n") == 1:
+			z["-m"] = z["-n"]
 
 		opt = z
 		##
@@ -568,6 +583,8 @@ def process_input(avail_opt):
 				name = "module_name"
 			elif name == "action" and 1 == avail_opt.count("io_fencing"):
 				name = "io_fencing"
+			elif name == "port" and 1 == avail_opt.count("drac_version"):
+				name = "module_name"
 
 			##
 			######
@@ -593,7 +610,7 @@ def check_input(device_opt, opt):
 	##
 	## Add options which are available for every fence agent
 	#####
-	device_opt.extend(common_opt)
+	device_opt.extend([x for x in common_opt if device_opt.count(x) == 0])
 	
 	options = dict(opt)
 	options["device_opt"] = device_opt
@@ -735,6 +752,8 @@ def fence_action(tn, options, set_power_fn, get_power_fn, get_outlet_list = None
 				print o + options["-C"] + alias	
 		return
 
+	if options["-o"] in ["off", "reboot"]:
+		time.sleep(int(options["-f"]))
 	status = get_power_fn(tn, options)
 
 	if status != "on" and status != "off":  
