@@ -19,6 +19,7 @@
 #include <string.h>
 #include <time.h>
 #include <sys/time.h>
+#include <errno.h>
 
 #include <linux/types.h>
 #include <linux/gfs2_ondisk.h>
@@ -59,6 +60,8 @@
 #define GFS_FORMAT_FS           (1309) /* Filesystem (all-encompassing) */
 #define GFS_FORMAT_MULTI        (1401) /* Multi-Host */
 
+#define DIV_RU(x, y) (((x) + (y) - 1) / (y))
+
 struct gfs1_rgrp {
 	struct gfs2_meta_header rg_header; /* hasn't changed from gfs1 to 2 */
 	uint32_t rg_flags;
@@ -110,6 +113,12 @@ struct gfs1_sb {
 	char sb_reserved[96];
 };
 
+struct inode_dir_block {
+	osi_list_t list;
+	uint64_t di_addr;
+	uint64_t di_paddr; /* Parent dir inode addr */
+};
+
 struct inode_block {
 	osi_list_t list;
 	uint64_t di_addr;
@@ -127,9 +136,11 @@ struct gfs1_sb  raw_gfs1_ondisk_sb;
 struct gfs2_sbd sb2;
 char device[256];
 struct inode_block dirs_to_fix;  /* linked list of directories to fix */
+struct inode_dir_block cdpns_to_fix; /* linked list of cdpn symlinks */
 int seconds;
 struct timeval tv;
 uint64_t dirs_fixed;
+uint64_t cdpns_fixed;
 uint64_t dirents_fixed;
 struct gfs1_jindex *sd_jindex = NULL;    /* gfs1 journal index in memory */
 int gfs2_inptrs;
@@ -313,8 +324,8 @@ static void fix_metatree(struct gfs2_sbd *sbp, struct gfs2_inode *ip,
 		bh = ip->i_bh;
 		/* First, build up the metatree */
 		for (h = 0; h < blk->height; h++) {
-			lookup_block(ip, ip->i_bh, h, &blk->mp, 1, &new,
-				     &block);
+			new = 0;
+			lookup_block(ip, bh, h, &blk->mp, 1, &new, &block);
 			if (bh != ip->i_bh)
 				brelse(bh);
 			if (!block)
@@ -633,8 +644,6 @@ static void fix_jdatatree(struct gfs2_sbd *sbp, struct gfs2_inode *ip,
 	mh.mh_type = GFS2_METATYPE_IN;
 	mh.mh_format = GFS2_FORMAT_IN;
 
-	/* This condition should never arise. 
-	   We're always dealing with unstuffed inodes */
 	if (!ip->i_di.di_height)
 		unstuff_dinode(ip);
 
@@ -645,8 +654,8 @@ static void fix_jdatatree(struct gfs2_sbd *sbp, struct gfs2_inode *ip,
 		bh = ip->i_bh;
 		/* First, build up the metatree */
 		for (h = 0; h < blk->height; h++) {
-			lookup_block(ip, ip->i_bh, h, &blk->mp, 1, &new,
-				     &block);
+			new = 0;
+			lookup_block(ip, bh, h, &blk->mp, 1, &new, &block);
 			if (bh != ip->i_bh)
 				brelse(bh);
 			if (!block)
@@ -739,16 +748,14 @@ static int adjust_jdata_inode(struct gfs2_sbd *sbp, struct gfs2_inode *ip)
 	       sbp->bsize - sizeof(struct gfs_dinode));
 	osi_list_add_prev(&blk->list, &blocks.list);
 
-	/* Now run the metadata chain and build lists of all data/metadata blocks */
+	/* Now run the metadata chain and build lists of all metadata blocks */
 	osi_list_foreach(tmp, &blocks.list) {
 		blk = osi_list_entry(tmp, struct blocklist, list);
 
-		if (blk->height >= ip->i_di.di_height)
+		if (blk->height >= ip->i_di.di_height - 1)
 			continue;
-
 		header_size = (blk->height > 0 ? sizeof(struct gfs_indirect) :
 			       sizeof(struct gfs_dinode));
-
 		for (ptr1 = (uint64_t *)blk->ptrbuf, ptrnum = 0;
 		     ptrnum < sbp->sd_inptrs; ptr1++, ptrnum++) {
 			if (!*ptr1)
@@ -781,23 +788,13 @@ static int adjust_jdata_inode(struct gfs2_sbd *sbp, struct gfs2_inode *ip)
 			/* Queue it to be processed later on in the loop. */
 			osi_list_add_prev(&newblk->list, &blocks.list);
 
+			/* read the new metadata block's pointers */
 			bh = bread(sbp, block);
-			if (newblk->height == ip->i_di.di_height) {
-				/* read in the jdata block */
-				memcpy(newblk->ptrbuf, bh->b_data +
-				       sizeof(struct gfs2_meta_header), bufsize);
-				/* Zero the buffer so we can fill it in later */
-				memset(bh->b_data + sizeof(struct gfs2_meta_header), 0,
-				       bufsize);
-			} else {
-				/* read the new metadata block's pointers */
-				memcpy(newblk->ptrbuf, bh->b_data +
-				       sizeof(struct gfs_indirect),
-				       sbp->bsize - sizeof(struct gfs_indirect));
-				/* Zero the buffer so we can fill it in later */
-				memset(bh->b_data + sizeof(struct gfs_indirect), 0,
-				       sbp->bsize - sizeof(struct gfs_indirect));
-			}
+			memcpy(newblk->ptrbuf, bh->b_data + sizeof(struct gfs_indirect),
+			       sbp->bsize - sizeof(struct gfs_indirect));
+			/* Zero the buffer so we can fill it in later */
+			memset(bh->b_data + sizeof(struct gfs_indirect), 0,
+			       sbp->bsize - sizeof(struct gfs_indirect));
 			bmodified(bh);
 			brelse(bh);
 			/* Free the block so we can reuse it. This allows us to
@@ -822,19 +819,69 @@ static int adjust_jdata_inode(struct gfs2_sbd *sbp, struct gfs2_inode *ip)
 		unsigned int len;
 
 		blk = osi_list_entry(tmp, struct blocklist, list);
-		/* If it's not a data block at the highest level */
-		if (blk->height != di_height) {
+		/* If it's not a highest level indirect block */
+		if (blk->height != di_height - 1) {
 			osi_list_del(tmp);
 			free(blk->ptrbuf);
 			free(blk);
 			continue;
 		}
-		len = bufsize;
-		jdata_mp_gfs1_to_gfs2(sbp, di_height, gfs2_hgt, &blk->mp, &gfs2mp, &len, dinode_size);
-		memcpy(&blk->mp, &gfs2mp, sizeof(struct metapath));
-		blk->height -= di_height - gfs2_hgt;
-		if (len)
-			fix_jdatatree(sbp, ip, blk, blk->ptrbuf, len);
+		/*
+		 * For each metadata block that holds jdata block pointers,
+		 * get the blk pointers and copy them block by block
+		 */
+		for (ptr1 = (uint64_t *) blk->ptrbuf, ptrnum = 0;
+		     ptrnum < sbp->sd_inptrs; ptr1++, ptrnum++) {
+			if (!*ptr1)
+				continue;
+			block = be64_to_cpu(*ptr1);
+
+			newblk = malloc(sizeof(struct blocklist));
+			if (!newblk) {
+				log_crit("Error: Can't allocate memory"
+					 " for indirect block fix.\n");
+				error = -1;
+				goto out;
+			}
+			memset(newblk, 0, sizeof(*newblk));
+			newblk->ptrbuf = malloc(bufsize); 
+			if (!newblk->ptrbuf) {
+				log_crit("Error: Can't allocate memory"
+					 " for file conversion.\n");
+				free(newblk);
+				goto out;
+			}
+			memset(newblk->ptrbuf, 0, bufsize);
+			newblk->block = block;
+			newblk->height = blk->height + 1;
+			/* Build the metapointer list from our predecessors */
+			for (h=0; h < blk->height; h++)
+				newblk->mp.mp_list[h] = blk->mp.mp_list[h];
+			newblk->mp.mp_list[h] = ptrnum;
+			bh = bread(sbp, block);
+			/* This is a data block. i.e newblk->height == ip->i_di.di_height */
+			/* read in the jdata block */
+			memcpy(newblk->ptrbuf, bh->b_data +
+			       sizeof(struct gfs2_meta_header), bufsize);
+			memset(bh->b_data + sizeof(struct gfs2_meta_header), 0,
+			       bufsize);
+			bmodified(bh);
+			brelse(bh);
+			/* Free the block so we can reuse it. This allows us to
+			   convert a "full" file system */
+			ip->i_di.di_blocks--;
+			gfs2_free_block(sbp, block);
+
+			len = bufsize;
+			jdata_mp_gfs1_to_gfs2(sbp, di_height, gfs2_hgt, &newblk->mp, &gfs2mp,
+					      &len, dinode_size);
+			memcpy(&newblk->mp, &gfs2mp, sizeof(struct metapath));
+			newblk->height -= di_height - gfs2_hgt;
+			if (len)
+				fix_jdatatree(sbp, ip, newblk, newblk->ptrbuf, len);
+			free(newblk->ptrbuf);
+			free(newblk);
+		}
 		osi_list_del(tmp);
 		free(blk->ptrbuf);
 		free(blk);
@@ -856,6 +903,81 @@ out:
 	return error;
 }
 
+const char *cdpn[14] = {"{hostname}", "{mach}", "{os}", "{uid}", "{gid}", "{sys}", "{jid}",
+			"@hostname", "@mach", "@os", "@uid", "@gid", "@sys", "@jid"};
+static int has_cdpn(const char *str)
+{
+	int i;
+	for (i=0; i<14; i++)
+		if (strstr(str, cdpn[i]) != NULL)
+			return 1;
+	return 0;
+}
+
+static int fix_cdpn_symlink(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh, struct gfs2_inode *ip)
+{
+	int ret = 0;
+	char *linkptr = NULL;
+
+	if (ip->i_di.di_height != 0)
+		return 0;
+
+	linkptr = bh->b_data + sizeof(struct gfs_dinode);
+	if (has_cdpn(linkptr)) {
+		struct inode_dir_block *fix;
+		/* Save the symlink di_addr. We'll find the parent di_addr later */
+		fix = malloc(sizeof(struct inode_dir_block));
+		if (!fix) {
+			log_crit("Error: out of memory.\n");
+			return -1;
+		}
+		memset(fix, 0, sizeof(struct inode_dir_block));
+		fix->di_addr = ip->i_di.di_num.no_addr;
+		osi_list_add_prev((osi_list_t *)&fix->list,
+				  (osi_list_t *)&cdpns_to_fix);
+	}
+
+	return ret;
+}
+
+/*
+ * fix_xattr -
+ * Extended attributes can be either direct (in the ip->i_di.di_eattr block) or
+ * then can be at a maximum of 1 indirect level. Multiple levels of indirection
+ * are not supported. If the di_eattr block contains extended attribute data,
+ * i.e block type = GFS_METATYPE_EA, we ignore it.
+ * If the di_eattr block contains block pointers to extended attributes we need
+ * to fix the header. gfs1 uses gfs_indirect as the header which is 64 bytes
+ * bigger than gfs2_meta_header that gfs2 uses.
+ */
+static int fix_xattr(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh, struct gfs2_inode *ip)
+{
+	int ret = 0, len, old_hdr_sz, new_hdr_sz;
+	struct gfs2_buffer_head *eabh;
+	char *buf;
+
+	/* Read in the i_di.di_eattr block */
+	eabh = bread(sbp, ip->i_di.di_eattr);
+        if (!gfs2_check_meta(eabh, GFS_METATYPE_IN)) {/* if it is an indirect block */
+		len = sbp->bsize - sizeof(struct gfs_indirect);
+		buf = malloc(len);
+		if (!buf) {
+			log_crit("Error: out of memory.\n");
+			return -1;
+		}
+		old_hdr_sz = sizeof(struct gfs_indirect);
+		new_hdr_sz = sizeof(struct gfs2_meta_header);
+		memcpy(buf, eabh->b_data + old_hdr_sz, sbp->bsize - old_hdr_sz);
+		memset(eabh->b_data + new_hdr_sz, 0, sbp->bsize - new_hdr_sz);
+		memcpy(eabh->b_data + new_hdr_sz, buf, len);
+		free(buf);
+		bmodified(eabh);
+	}
+        brelse(eabh);
+
+	return ret;
+}
+
 /* ------------------------------------------------------------------------- */
 /* adjust_inode - change an inode from gfs1 to gfs2                          */
 /*                                                                           */
@@ -941,6 +1063,18 @@ static int adjust_inode(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh)
 			ret = adjust_indirect_blocks(sbp, inode);
 		if (ret)
 			return -1;
+		/* Check for cdpns */
+		if (inode->i_di.di_mode & S_IFLNK) {
+			ret = fix_cdpn_symlink(sbp, bh, inode);
+			if (ret)
+				return -1;
+		}
+		/* Check for extended attributes */
+		if (inode->i_di.di_eattr) {
+			ret = fix_xattr(sbp, bh, inode);
+			if (ret)
+				return -1;
+		}
 	}
 	
 	bmodified(inode->i_bh);
@@ -957,7 +1091,7 @@ static int adjust_inode(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh)
 /*                                                                           */
 /* Returns: 0 on success, -1 on failure                                      */
 /* ------------------------------------------------------------------------- */
-static int inode_renumber(struct gfs2_sbd *sbp, uint64_t root_inode_addr)
+static int inode_renumber(struct gfs2_sbd *sbp, uint64_t root_inode_addr, osi_list_t *cdpn_to_fix)
 {
 	struct rgrp_list *rgd;
 	osi_list_t *tmp;
@@ -1045,13 +1179,16 @@ static int inode_renumber(struct gfs2_sbd *sbp, uint64_t root_inode_addr)
 /* fetch_inum - fetch an inum entry from disk, given its block               */
 /* ------------------------------------------------------------------------- */
 static int fetch_inum(struct gfs2_sbd *sbp, uint64_t iblock,
-					   struct gfs2_inum *inum)
+		      struct gfs2_inum *inum, uint64_t *eablk)
 {
 	struct gfs2_inode *fix_inode;
 
 	fix_inode = inode_read(sbp, iblock);
 	inum->no_formal_ino = fix_inode->i_di.di_num.no_formal_ino;
 	inum->no_addr = fix_inode->i_di.di_num.no_addr;
+	if (eablk)
+		*eablk = fix_inode->i_di.di_eattr;
+
 	inode_put(&fix_inode);
 	return 0;
 }/* fetch_inum */
@@ -1062,12 +1199,12 @@ static int fetch_inum(struct gfs2_sbd *sbp, uint64_t iblock,
 /* We changed inode numbers, so we must update that number into the          */
 /* directory entries themselves.                                             */
 /*                                                                           */
-/* Returns: 0 on success, -1 on failure                                      */
+/* Returns: 0 on success, -1 on failure, -EISDIR when dentmod marked DT_DIR  */
 /* ------------------------------------------------------------------------- */
 static int process_dirent_info(struct gfs2_inode *dip, struct gfs2_sbd *sbp,
-			       struct gfs2_buffer_head *bh, int dir_entries)
+			       struct gfs2_buffer_head *bh, int dir_entries, uint64_t dentmod)
 {
-	int error;
+	int error = 0;
 	struct gfs2_dirent *dent;
 	int de; /* directory entry index */
 	
@@ -1076,12 +1213,23 @@ static int process_dirent_info(struct gfs2_inode *dip, struct gfs2_sbd *sbp,
 		log_crit("Error retrieving directory.\n");
 		return -1;
 	}
+	error = 0;
 	/* Go through every dirent in the buffer and process it. */
 	/* Turns out you can't trust dir_entries is correct.     */
 	for (de = 0; ; de++) {
 		struct gfs2_inum inum;
 		int dent_was_gfs1;
-		
+
+		if (dentmod) {
+			if (dent->de_type == cpu_to_be16(DT_LNK)
+			    && cpu_to_be64(dent->de_inum.no_addr) == dentmod) {
+				dent->de_type = cpu_to_be16(DT_DIR);
+				error = -EISDIR;
+				break;
+			}
+			goto skip_next;
+		}
+
 		gettimeofday(&tv, NULL);
 		/* Do more warm fuzzy stuff for the customer. */
 		dirents_fixed++;
@@ -1095,7 +1243,7 @@ static int process_dirent_info(struct gfs2_inode *dip, struct gfs2_sbd *sbp,
 		gfs2_inum_in(&inum, (char *)&dent->de_inum);
 		dent_was_gfs1 = (dent->de_inum.no_addr == dent->de_inum.no_formal_ino);
 		if (inum.no_formal_ino) { /* if not a sentinel (placeholder) */
-			error = fetch_inum(sbp, inum.no_addr, &inum);
+			error = fetch_inum(sbp, inum.no_addr, &inum, NULL);
 			if (error) {
 				log_crit("Error retrieving inode 0x%llx\n",
 					 (unsigned long long)inum.no_addr);
@@ -1139,11 +1287,30 @@ static int process_dirent_info(struct gfs2_inode *dip, struct gfs2_sbd *sbp,
 				break;
 			}
 		}
+		/*
+		 * Compare this dirent address with every one in the
+		 * cdpns_to_fix list to find if this directory (dip) is
+		 * a cdpn symlink's parent. If so add it to the list element
+		 */
+		if (dent->de_type == cpu_to_be16(DT_LNK)) {
+			osi_list_t *tmp;
+			struct inode_dir_block *fix;
+			osi_list_foreach(tmp, &cdpns_to_fix.list) {
+				fix = osi_list_entry(tmp, struct inode_dir_block, list);
+				if (fix->di_addr == inum.no_addr)
+					fix->di_paddr = dip->i_di.di_num.no_addr;
+			}
+		}
+
+	skip_next:
 		error = gfs2_dirent_next(dip, bh, &dent);
-		if (error)
+		if (error) {
+			if (error == -ENOENT) /* beyond the end of this bh */
+				error = 0;
 			break;
+		}
 	} /* for every directory entry */
-	return 0;
+	return error;
 }/* process_dirent_info */
 
 /* ------------------------------------------------------------------------- */
@@ -1154,7 +1321,7 @@ static int process_dirent_info(struct gfs2_inode *dip, struct gfs2_sbd *sbp,
 /*                                                                           */
 /* Returns: 0 on success, -1 on failure                                      */
 /* ------------------------------------------------------------------------- */
-static int fix_one_directory_exhash(struct gfs2_sbd *sbp, struct gfs2_inode *dip)
+static int fix_one_directory_exhash(struct gfs2_sbd *sbp, struct gfs2_inode *dip, uint64_t dentmod)
 {
 	struct gfs2_buffer_head *bh_leaf;
 	int error;
@@ -1190,13 +1357,40 @@ static int fix_one_directory_exhash(struct gfs2_sbd *sbp, struct gfs2_inode *dip
 			break;
 		}
 		gfs2_leaf_in(&leaf, bh_leaf); /* buffer to structure */
-		error = process_dirent_info(dip, sbp, bh_leaf, leaf.lf_entries);
+		error = process_dirent_info(dip, sbp, bh_leaf, leaf.lf_entries, dentmod);
 		bmodified(bh_leaf);
 		brelse(bh_leaf);
+		if (dentmod && error == -EISDIR) /* dentmod was marked DT_DIR, break out */
+			break;
 	} /* for leaf_num */
 	return 0;
 }/* fix_one_directory_exhash */
 
+static int process_directory(struct gfs2_sbd *sbp, uint64_t dirblock, uint64_t dentmod)
+{
+	struct gfs2_inode *dip;
+	int error = 0;
+	/* read in the directory inode */
+	dip = inode_read(sbp, dirblock);
+	/* fix the directory: either exhash (leaves) or linear (stuffed) */
+	if (dip->i_di.di_flags & GFS2_DIF_EXHASH) {
+		if (fix_one_directory_exhash(sbp, dip, dentmod)) {
+			log_crit("Error fixing exhash directory.\n");
+			inode_put(&dip);
+			return -1;
+		}
+	} else {
+		error = process_dirent_info(dip, sbp, dip->i_bh, dip->i_di.di_entries, dentmod);
+		if (error && error != -EISDIR) {
+			log_crit("Error fixing linear directory.\n");
+			inode_put(&dip);
+			return -1;
+		}
+	}
+	bmodified(dip->i_bh);
+	inode_put(&dip);
+	return 0;
+}
 /* ------------------------------------------------------------------------- */
 /* fix_directory_info - sync new inode numbers with directory info           */
 /* Returns: 0 on success, -1 on failure                                      */
@@ -1206,7 +1400,6 @@ static int fix_directory_info(struct gfs2_sbd *sbp, osi_list_t *dir_to_fix)
 	osi_list_t *tmp, *fix;
 	struct inode_block *dir_iblk;
 	uint64_t offset, dirblock;
-	struct gfs2_inode *dip;
 
 	dirs_fixed = 0;
 	dirents_fixed = 0;
@@ -1227,25 +1420,10 @@ static int fix_directory_info(struct gfs2_sbd *sbp, osi_list_t *dir_to_fix)
 		/* figure out the directory inode block and read it in */
 		dir_iblk = (struct inode_block *)fix;
 		dirblock = dir_iblk->di_addr; /* addr of dir inode */
-		/* read in the directory inode */
-		dip = inode_read(sbp, dirblock);
-		/* fix the directory: either exhash (leaves) or linear (stuffed) */
-		if (dip->i_di.di_flags & GFS2_DIF_EXHASH) {
-			if (fix_one_directory_exhash(sbp, dip)) {
-				log_crit("Error fixing exhash directory.\n");
-				inode_put(&dip);
-				return -1;
-			}
+		if (process_directory(sbp, dirblock, 0)) {
+			log_crit("Error processing directory\n");
+			return -1;
 		}
-		else {
-			if (process_dirent_info(dip, sbp, dip->i_bh,
-						dip->i_di.di_entries)) {
-				log_crit("Error fixing linear directory.\n");
-				inode_put(&dip);
-				return -1;
-			}
-		}
-		inode_put(&dip);
 	}
 	/* Free the last entry in memory: */
 	if (tmp) {
@@ -1255,6 +1433,60 @@ static int fix_directory_info(struct gfs2_sbd *sbp, osi_list_t *dir_to_fix)
 	return 0;
 }/* fix_directory_info */
 
+/* ------------------------------------------------------------------------- */
+/* fix_cdpn_symlinks - convert cdpn symlinks to empty directories            */
+/* Returns: 0 on success, -1 on failure                                      */
+/* ------------------------------------------------------------------------- */
+static int fix_cdpn_symlinks(struct gfs2_sbd *sbp, osi_list_t *cdpn_to_fix)
+{
+	osi_list_t *tmp, *x;
+	int error = 0;
+
+	cdpns_fixed = 0;
+	osi_list_foreach_safe(tmp, cdpn_to_fix, x) {
+		struct gfs2_inum fix, dir;
+		struct inode_dir_block *l_fix;
+		struct gfs2_buffer_head *bh;
+		struct gfs2_inode *fix_inode;
+		uint64_t eablk;
+
+		l_fix = osi_list_entry(tmp, struct inode_dir_block, list);
+		osi_list_del(tmp);
+
+		/* convert symlink to empty dir */
+		error = fetch_inum(sbp, l_fix->di_addr, &fix, &eablk);
+		if (error) {
+			log_crit("Error retrieving inode at block %llx\n", 
+				 (unsigned long long)l_fix->di_addr);
+			break;
+		}
+		error = fetch_inum(sbp, l_fix->di_paddr, &dir, NULL);
+		if (error) {
+			log_crit("Error retrieving inode at block %llx\n",
+				 (unsigned long long)l_fix->di_paddr);
+			break;
+		}
+
+		/* initialize the symlink inode to be a directory */
+		bh = init_dinode(sbp, &fix, S_IFDIR | 0755, 0, &dir);
+		fix_inode = inode_get(sbp, bh);
+		fix_inode->i_di.di_eattr = eablk; /*fix extended attribute */
+		inode_put(&fix_inode);
+		bmodified(bh);
+		brelse(bh);
+
+		/* fix the parent directory dirent entry for this inode */
+		error = process_directory(sbp, l_fix->di_paddr, l_fix->di_addr);
+		if (error) {
+			log_crit("Error trying to fix cdpn dentry\n");
+			break;
+		}
+		free(l_fix);
+		cdpns_fixed++;
+	}
+	return error;
+} /* fix_cdpn_symlinks */
+
 /* ------------------------------------------------------------------------- */
 /* Fetch gfs1 jindex structure from buffer                                   */
 /* ------------------------------------------------------------------------- */
@@ -1277,6 +1509,7 @@ static int read_gfs1_jiindex(struct gfs2_sbd *sdp)
 	char buf[sizeof(struct gfs1_jindex)];
 	unsigned int j;
 	int error=0;
+	unsigned int tmp_mode = 0;
 
 	if(ip->i_di.di_size % sizeof(struct gfs1_jindex) != 0){
 		log_crit("The size reported in the journal index"
@@ -1292,6 +1525,14 @@ static int read_gfs1_jiindex(struct gfs2_sbd *sdp)
 		log_crit("Unable to zero journal index\n");
 		return -1;
 	}
+	/* ugly hack
+	 * Faking the gfs1_jindex inode as a directory to gfs2_readi
+	 * so it skips the metaheader struct in the data blocks
+	 * in the inode. gfs2_jindex inode doesn't have metaheaders
+	 * in the data blocks */
+	tmp_mode = ip->i_di.di_mode;
+	ip->i_di.di_mode &= ~S_IFMT;
+	ip->i_di.di_mode |= S_IFDIR;
 	for (j = 0; ; j++) {
 		struct gfs1_jindex *journ;
 
@@ -1308,6 +1549,7 @@ static int read_gfs1_jiindex(struct gfs2_sbd *sdp)
 		gfs1_jindex_in(journ, buf);
 		sdp->jsize = (journ->ji_nsegment * 16 * sdp->bsize) >> 20;
 	}
+	ip->i_di.di_mode = tmp_mode;
 	if(j * sizeof(struct gfs1_jindex) != ip->i_di.di_size){
 		log_crit("journal inode size invalid\n");
 		goto fail;
@@ -1343,6 +1585,7 @@ static int init(struct gfs2_sbd *sbp)
 	sbp->sd_sb.sb_header.mh_format = GFS2_FORMAT_SB;
 
 	osi_list_init((osi_list_t *)&dirs_to_fix);
+	osi_list_init((osi_list_t *)&cdpns_to_fix);
 	/* ---------------------------------------------- */
 	/* Initialize lists and read in the superblock.   */
 	/* ---------------------------------------------- */
@@ -1744,22 +1987,24 @@ static void conv_build_jindex(struct gfs2_sbd *sdp)
 	sdp->md.jiinode = createi(sdp->master_dir, "jindex", S_IFDIR | 0700,
 				  GFS2_DIF_SYSTEM);
 
+	sdp->md.journal = malloc(sdp->md.journals *
+				 sizeof(struct gfs2_inode *));
 	for (j = 0; j < sdp->md.journals; j++) {
 		char name[256];
-		struct gfs2_inode *ip;
 
 		printf("Writing journal #%d...", j + 1);
 		fflush(stdout);
 		sprintf(name, "journal%u", j);
-		ip = createi(sdp->md.jiinode, name, S_IFREG | 0600,
-			     GFS2_DIF_SYSTEM);
-		write_journal(sdp, ip, j,
+		sdp->md.journal[j] = createi(sdp->md.jiinode, name, S_IFREG |
+					     0600, GFS2_DIF_SYSTEM);
+		write_journal(sdp, j,
 			      sdp->jsize << 20 >> sdp->sd_sb.sb_bsize_shift);
-		inode_put(&ip);
+		inode_put(&sdp->md.journal[j]);
 		printf("done.\n");
 		fflush(stdout);
 	}
 
+	free(sdp->md.journal);
 	if (sdp->debug) {
 		printf("\nJindex:\n");
 		gfs2_dinode_print(&sdp->md.jiinode->i_di);
@@ -1768,6 +2013,134 @@ static void conv_build_jindex(struct gfs2_sbd *sdp)
 	inode_put(&sdp->md.jiinode);
 }
 
+static unsigned int total_file_blocks(struct gfs2_sbd *sdp, 
+				      uint64_t filesize, int journaled)
+{
+	unsigned int data_blks = 0, meta_blks = 0, total_blks;
+	unsigned int max, height, bsize;
+	uint64_t *arr;
+
+	/* Now find the total meta blocks required for data_blks */
+	if (filesize <= sdp->bsize - sizeof(struct gfs2_dinode)) {
+		total_blks = 1; /* stuffed inode */
+		goto out;
+	}
+
+	if (journaled) {
+		arr = sdp->sd_jheightsize;
+		max = sdp->sd_max_jheight;
+		bsize = sdp->sd_jbsize;
+	} else {
+		arr = sdp->sd_heightsize;
+		max = sdp->sd_max_height;
+		bsize = sdp->bsize;
+	}
+	data_blks = DIV_RU(filesize, bsize); /* total data blocks reqd */
+
+	for (height = 0; height < max; height++)
+		if (arr[height] >= filesize)
+			break;
+	if (height == 1) {
+		total_blks = data_blks + 1; /* dinode has direct ptrs to data blocks */
+		goto out;
+	}
+
+	meta_blks = DIV_RU(data_blks, sdp->sd_inptrs);
+	total_blks = data_blks + meta_blks;
+out:
+	return data_blks + meta_blks;
+}
+
+/* We check if the GFS2 filesystem files/structures created after the call to 
+ * check_fit() in main() will fit in the currently available free blocks
+ */
+static int check_fit(struct gfs2_sbd *sdp)
+{
+	unsigned int blks_need = 0, blks_avail = sdp->blks_total - sdp->blks_alloced;
+
+	/* build_master() */
+	blks_need++; /*creation of master dir inode - 1 block */
+
+	/* conv_build_jindex() */
+	{
+		blks_need++; /* creation of 'jindex' disk inode */
+		/* creation of journals */
+		blks_need += sdp->md.journals *
+			total_file_blocks(sdp, sdp->jsize << 20, 1);
+	}
+	/* build_per_node() */
+	{
+		blks_need++; /* creation of 'per_node' dir inode */
+		/* njourn x (inum_range + statfs_change + quota_change inodes) */
+		blks_need += sdp->md.journals * 3;
+		/* quota change inodes are prealloced */
+		blks_need += sdp->md.journals *
+			total_file_blocks(sdp, sdp->qcsize << 20, 1);
+	}
+	/* build_inum() */
+	blks_need++; /* creation of 'inum' disk inode */
+
+	/* build_statfs() */
+	blks_need++; /* creation of 'statfs' disk inode */
+
+	/* build_rindex() */
+	{
+		osi_list_t *tmp, *head;
+		unsigned int rg_count = 0;
+
+		blks_need++; /* creationg of 'rindex' disk inode */
+		/* find the total # of rindex entries, gives size of rindex inode */
+		for (head = &sdp->rglist, tmp = head->next; tmp != head;
+		     tmp = tmp->next)
+			rg_count++;
+		blks_need += 
+			total_file_blocks(sdp, rg_count * sizeof(struct gfs2_rindex), 1);
+	}
+	/* build_quota() */
+	blks_need++; /* quota inode block and uid=gid=0 quota - total 1 block */
+
+	/* Up until this point we require blks_need blocks. We don't 
+	 * include the blocks freed by the next step (remove_obsolete_gfs1)
+	 * because it's possible for us to exceed the available blocks
+	 * before this step */
+
+	return blks_avail > blks_need;
+}
+
+/* We fetch the old quota inode block and copy the contents of the block
+ * (minus the struct gfs2_dinode) into the new quota block. We update the 
+ * inode height/size of the new quota file to that of the old one and set the 
+ * old quota inode height/size to zero, so only the inode block gets freed.
+ */
+static void copy_quotas(struct gfs2_sbd *sdp)
+{
+	struct gfs2_inum inum;
+	struct gfs2_inode *oq_ip, *nq_ip;
+	int err;
+
+	err = gfs2_lookupi(sdp->master_dir, "quota", 5, &nq_ip);
+	if (err)
+		die("Couldn't lookup new quota file: %d\n", err);
+
+	gfs2_inum_in(&inum, (char *)&raw_gfs1_ondisk_sb.sb_quota_di);
+	oq_ip = inode_read(sdp, inum.no_addr);
+
+	nq_ip->i_di.di_height = oq_ip->i_di.di_height;
+	nq_ip->i_di.di_size = oq_ip->i_di.di_size;
+	memcpy(nq_ip->i_bh->b_data + sizeof(struct gfs2_dinode), 
+	       oq_ip->i_bh->b_data + sizeof(struct gfs2_dinode),
+	       sdp->bsize - sizeof(struct gfs2_dinode));
+
+	oq_ip->i_di.di_height = 0;
+	oq_ip->i_di.di_size = 0;
+
+	bmodified(nq_ip->i_bh);
+	inode_put(&nq_ip);
+
+	bmodified(oq_ip->i_bh);
+	inode_put(&oq_ip);
+}
+
 /* ------------------------------------------------------------------------- */
 /* main - mainline code                                                      */
 /* ------------------------------------------------------------------------- */
@@ -1813,7 +2186,8 @@ int main(int argc, char **argv)
 	/* Renumber the inodes consecutively.             */
 	/* ---------------------------------------------- */
 	if (!error) {
-		error = inode_renumber(&sb2, sb2.sd_sb.sb_root_dir.no_addr);
+		error = inode_renumber(&sb2, sb2.sd_sb.sb_root_dir.no_addr,
+				       (osi_list_t *)&cdpns_to_fix);
 		if (error)
 			log_crit("\n%s: Error renumbering inodes.\n", device);
 		fsync(sb2.device_fd); /* write the buffers to disk */
@@ -1830,6 +2204,17 @@ int main(int argc, char **argv)
 			log_crit("\n%s: Error fixing directories.\n", device);
 	}
 	/* ---------------------------------------------- */
+	/* Convert cdpn symlinks to empty dirs            */
+	/* ---------------------------------------------- */
+	if (!error) {
+		error = fix_cdpn_symlinks(&sb2, (osi_list_t *)&cdpns_to_fix);
+		log_notice("\r%" PRIu64 " cdpn symlinks moved to empty directories.",
+			   cdpns_fixed);
+		fflush(stdout);
+		if (error)
+			log_crit("\n%s: Error fixing cdpn symlinks.\n", device);
+	}
+	/* ---------------------------------------------- */
 	/* Convert journal space to rg space              */
 	/* ---------------------------------------------- */
 	if (!error) {
@@ -1843,12 +2228,22 @@ int main(int argc, char **argv)
 	/* Create our system files and directories.       */
 	/* ---------------------------------------------- */
 	if (!error) {
+		int jreduce = 0;
 		/* Now we've got to treat it as a gfs2 file system */
 		if (compute_constants(&sb2)) {
 			log_crit("Error: Bad constants (1)\n");
 			exit(-1);
 		}
 
+		/* Check if all the files we're about to create will 
+		 * fit into the space remaining on the device */
+		while (!check_fit(&sb2)) {
+			sb2.jsize--; /* reduce jsize by 1MB each time */
+			jreduce = 1;
+		}
+		if (jreduce)
+			log_notice("Reduced journal size to %u MB to accommodate "
+				   "GFS2 file system structures.\n", sb2.jsize);
 		/* Build the master subdirectory. */
 		build_master(&sb2); /* Does not do inode_put */
 		sb2.sd_sb.sb_master_dir = sb2.master_dir->i_di.di_num;
@@ -1867,6 +2262,15 @@ int main(int argc, char **argv)
 		/* Create the quota file */
 		build_quota(&sb2);
 
+		/* Copy out the master dinode */
+		{
+			struct gfs2_inode *ip = sb2.master_dir;
+			if (ip->i_bh->b_modified)
+				gfs2_dinode_out(&ip->i_di, ip->i_bh);
+		}
+		/* Copy old quotas */
+		copy_quotas(&sb2);
+
 		update_inode_file(&sb2);
 		/* Now delete the now-obsolete gfs1 files: */
 		remove_obsolete_gfs1(&sb2);
