@@ -35,9 +35,10 @@ struct block_count {
 static int leaf(struct gfs2_inode *ip, uint64_t block,
 		struct gfs2_buffer_head *bh, void *private);
 static int check_metalist(struct gfs2_inode *ip, uint64_t block,
-			  struct gfs2_buffer_head **bh, void *private);
+			  struct gfs2_buffer_head **bh, int h, void *private);
 static int undo_check_metalist(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, void *private);
+			       struct gfs2_buffer_head **bh, int h,
+			       void *private);
 static int check_data(struct gfs2_inode *ip, uint64_t block, void *private);
 static int undo_check_data(struct gfs2_inode *ip, uint64_t block,
 			   void *private);
@@ -60,7 +61,8 @@ static int check_extended_leaf_eattr(struct gfs2_inode *ip, uint64_t *data_ptr,
 static int finish_eattr_indir(struct gfs2_inode *ip, int leaf_pointers,
 			      int leaf_pointer_errors, void *private);
 static int invalidate_metadata(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, void *private);
+			       struct gfs2_buffer_head **bh, int h,
+			       void *private);
 static int invalidate_leaf(struct gfs2_inode *ip, uint64_t block,
 			   struct gfs2_buffer_head *bh, void *private);
 static int invalidate_data(struct gfs2_inode *ip, uint64_t block,
@@ -72,6 +74,7 @@ static int invalidate_eattr_indir(struct gfs2_inode *ip, uint64_t block,
 static int invalidate_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 				 uint64_t parent, struct gfs2_buffer_head **bh,
 				 void *private);
+static int handle_ip(struct gfs2_sbd *sdp, struct gfs2_inode *ip);
 
 struct metawalk_fxns pass1_fxns = {
 	.private = NULL,
@@ -101,6 +104,99 @@ struct metawalk_fxns invalidate_fxns = {
 	.check_eattr_leaf = invalidate_eattr_leaf,
 };
 
+/*
+ * resuscitate_metalist - make sure a system directory entry's metadata blocks
+ *                        are marked "in use" in the bitmap.
+ *
+ * This function makes sure metadata blocks for system and root directories are
+ * marked "in use" by the bitmap.  You don't want root's indirect blocks
+ * deleted, do you? Or worse, reused for lost+found.
+ */
+static int resuscitate_metalist(struct gfs2_inode *ip, uint64_t block,
+				struct gfs2_buffer_head **bh, int h,
+				void *private)
+{
+	struct block_count *bc = (struct block_count *)private;
+
+	*bh = NULL;
+	if (gfs2_check_range(ip->i_sbd, block)){ /* blk outside of FS */
+		fsck_blockmap_set(ip, ip->i_di.di_num.no_addr,
+				  _("itself"), gfs2_bad_block);
+		log_err( _("Bad indirect block pointer (out of range) "
+			   "found in system inode %lld (0x%llx).\n"),
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr);
+		return 1;
+	}
+	if (fsck_system_inode(ip->i_sbd, block))
+		fsck_blockmap_set(ip, block, _("system file"), gfs2_indir_blk);
+	else
+		check_n_fix_bitmap(ip->i_sbd, block, gfs2_indir_blk);
+	bc->indir_count++;
+	return 0;
+}
+
+/*
+ * resuscitate_dentry - make sure a system directory entry is alive
+ *
+ * This function makes sure directory entries in system directories are
+ * kept alive.  You don't want journal0 deleted from jindex, do you?
+ */
+static int resuscitate_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
+			      struct gfs2_dirent *prev_de,
+			      struct gfs2_buffer_head *bh, char *filename,
+			      uint16_t *count, void *priv)
+{
+	struct gfs2_sbd *sdp = ip->i_sbd;
+	struct gfs2_dirent dentry, *de;
+	char tmp_name[PATH_MAX];
+	uint64_t block;
+	enum gfs2_mark_block dinode_type;
+
+	memset(&dentry, 0, sizeof(struct gfs2_dirent));
+	gfs2_dirent_in(&dentry, (char *)dent);
+	de = &dentry;
+	block = de->de_inum.no_addr;
+	/* Start of checks */
+	memset(tmp_name, 0, sizeof(tmp_name));
+	if(de->de_name_len < sizeof(tmp_name))
+		strncpy(tmp_name, filename, de->de_name_len);
+	else
+		strncpy(tmp_name, filename, sizeof(tmp_name) - 1);
+	if(gfs2_check_range(sdp, block)) {
+		log_err( _("Block # referenced by system directory entry %s "
+			   "in inode %lld (0x%llx) is out of range; "
+			   "ignored.\n"),
+			 tmp_name, (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr);
+		return 0;
+	}
+	if (block == sdp->md.jiinode->i_di.di_num.no_addr ||
+	    block == sdp->md.pinode->i_di.di_num.no_addr ||
+	    block == sdp->master_dir->i_di.di_num.no_addr)
+		dinode_type = gfs2_inode_dir;
+	else
+		dinode_type = gfs2_inode_file;
+	/* If this is a system dinode, we'll handle it later in
+	   check_system_inodes.  If not, it'll be handled by pass1 but
+	   since it's in a system directory we need to make sure it's
+	   represented in the rgrp bitmap. */
+	if (fsck_system_inode(sdp, block))
+		fsck_blockmap_set(ip, block, _("system file"), dinode_type);
+	else
+		check_n_fix_bitmap(sdp, block, dinode_type);
+	/* Return the number of leaf entries so metawalk doesn't flag this
+	   leaf as having none. */
+	*count = be16_to_cpu(((struct gfs2_leaf *)bh->b_data)->lf_entries);
+	return 0;
+}
+
+struct metawalk_fxns sysdir_fxns = {
+	.private = NULL,
+	.check_metalist = resuscitate_metalist,
+	.check_dentry = resuscitate_dentry,
+};
+
 static int leaf(struct gfs2_inode *ip, uint64_t block,
 		struct gfs2_buffer_head *bh, void *private)
 {
@@ -112,7 +208,7 @@ static int leaf(struct gfs2_inode *ip, uint64_t block,
 }
 
 static int check_metalist(struct gfs2_inode *ip, uint64_t block,
-			  struct gfs2_buffer_head **bh, void *private)
+			  struct gfs2_buffer_head **bh, int h, void *private)
 {
 	uint8_t q;
 	int found_dup = 0, iblk_type;
@@ -132,7 +228,7 @@ static int check_metalist(struct gfs2_inode *ip, uint64_t block,
 
 		return 1;
 	}
-	if (S_ISDIR(ip->i_di.di_mode)) {
+	if (S_ISDIR(ip->i_di.di_mode) && h == ip->i_di.di_height) {
 		iblk_type = GFS2_METATYPE_JD;
 		blktypedesc = _("a directory hash table block");
 	} else {
@@ -155,13 +251,13 @@ static int check_metalist(struct gfs2_inode *ip, uint64_t block,
 	nbh = bread(ip->i_sbd, block);
 
 	if (gfs2_check_meta(nbh, iblk_type)){
-		log_debug( _("Inode %lld (0x%llx) has a bad indirect block "
-			     "pointer %lld (0x%llx) (points to something "
-			     "that is not %s).\n"),
-			   (unsigned long long)ip->i_di.di_num.no_addr,
-			   (unsigned long long)ip->i_di.di_num.no_addr,
-			   (unsigned long long)block,
-			   (unsigned long long)block, blktypedesc);
+		log_err( _("Inode %lld (0x%llx) has a bad indirect block "
+			   "pointer %lld (0x%llx) (points to something "
+			   "that is not %s).\n"),
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)block,
+			 (unsigned long long)block, blktypedesc);
 		if(!found_dup) {
 			fsck_blockmap_set(ip, block, _("bad indirect"),
 					  gfs2_meta_inval);
@@ -183,7 +279,8 @@ static int check_metalist(struct gfs2_inode *ip, uint64_t block,
 }
 
 static int undo_check_metalist(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, void *private)
+			       struct gfs2_buffer_head **bh, int h,
+			       void *private)
 {
 	struct duptree *d;
 	int found_dup = 0, iblk_type;
@@ -197,7 +294,7 @@ static int undo_check_metalist(struct gfs2_inode *ip, uint64_t block,
 				  _("itself"), gfs2_block_free);
 		return 1;
 	}
-	if (S_ISDIR(ip->i_di.di_mode))
+	if (S_ISDIR(ip->i_di.di_mode) && h == ip->i_di.di_height)
 		iblk_type = GFS2_METATYPE_JD;
 	else
 		iblk_type = GFS2_METATYPE_IN;
@@ -741,7 +838,8 @@ static int mark_block_invalid(struct gfs2_inode *ip, uint64_t block,
 }
 
 static int invalidate_metadata(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, void *private)
+			       struct gfs2_buffer_head **bh, int h,
+			       void *private)
 {
 	return mark_block_invalid(ip, block, ref_as_meta, _("metadata"));
 }
@@ -822,7 +920,8 @@ static int rangecheck_block(struct gfs2_inode *ip, uint64_t block,
 }
 
 static int rangecheck_metadata(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, void *private)
+			       struct gfs2_buffer_head **bh, int h,
+			       void *private)
 {
 	return rangecheck_block(ip, block, bh, _("metadata"), private);
 }
@@ -865,40 +964,15 @@ struct metawalk_fxns rangecheck_fxns = {
         .check_eattr_leaf = rangecheck_eattr_leaf,
 };
 
-static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
+/*
+ * handle_ip - process an incore structure representing a dinode.
+ */
+static int handle_ip(struct gfs2_sbd *sdp, struct gfs2_inode *ip)
 {
-	uint8_t q;
-	struct gfs2_inode *ip;
 	int error;
 	struct block_count bc = {0};
-	uint64_t block = bh->b_blocknr;
 	long bad_pointers;
-
-	q = block_type(block);
-	if(q != gfs2_block_free) {
-		log_err( _("Found a duplicate inode block at #%" PRIu64
-			   " (0x%" PRIx64 ") previously marked as a %s\n"),
-			 block, block, block_type_string(q));
-		add_duplicate_ref(ip, block, ref_as_meta, 0, INODE_VALID);
-		return 0;
-	}
-
-	ip = fsck_inode_get(sdp, bh);
-	if (ip->i_di.di_num.no_addr != block) {
-		log_err( _("Inode #%llu (0x%llx): Bad inode address found: %llu "
-			"(0x%llx)\n"), (unsigned long long)block,
-			(unsigned long long)block,
-			(unsigned long long)ip->i_di.di_num.no_addr,
-			(unsigned long long)ip->i_di.di_num.no_addr);
-		if(query( _("Fix address in inode at block #%"
-			    PRIu64 " (0x%" PRIx64 ")? (y/n) "),
-			 block, block)) {
-			ip->i_di.di_num.no_addr = ip->i_di.di_num.no_formal_ino = block;
-			bmodified(ip->i_bh);
-		} else
-			log_err( _("Address in inode at block #%" PRIu64
-				 " (0x%" PRIx64 ") not fixed\n"), block, block);
-	}
+	uint64_t block = ip->i_bh->b_blocknr;
 
 	bad_pointers = 0L;
 
@@ -916,7 +990,6 @@ static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
 			 BAD_POINTER_TOLERANCE);
 		fsck_blockmap_set(ip, ip->i_di.di_num.no_addr,
 				  _("badly corrupt"), gfs2_block_free);
-		fsck_inode_put(&ip);
 		return 0;
 	}
 
@@ -924,64 +997,40 @@ static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
 
 	case S_IFDIR:
 		if (fsck_blockmap_set(ip, block, _("directory"),
-				      gfs2_inode_dir)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
-		if(!dirtree_insert(block)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
+				      gfs2_inode_dir))
+			goto bad_dinode;
+		if(!dirtree_insert(block))
+			goto bad_dinode;
 		break;
 	case S_IFREG:
 		if (fsck_blockmap_set(ip, block, _("file"),
-				      gfs2_inode_file)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
+				      gfs2_inode_file))
+			goto bad_dinode;
 		break;
 	case S_IFLNK:
 		if (fsck_blockmap_set(ip, block, _("symlink"),
-				      gfs2_inode_lnk)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
+				      gfs2_inode_lnk))
+			goto bad_dinode;
 		break;
 	case S_IFBLK:
 		if (fsck_blockmap_set(ip, block, _("block device"),
-				      gfs2_inode_blk)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
+				      gfs2_inode_blk))
+			goto bad_dinode;
 		break;
 	case S_IFCHR:
 		if (fsck_blockmap_set(ip, block, _("character device"),
-				      gfs2_inode_chr)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
+				      gfs2_inode_chr))
+			goto bad_dinode;
 		break;
 	case S_IFIFO:
 		if (fsck_blockmap_set(ip, block, _("fifo"),
-				      gfs2_inode_fifo)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
+				      gfs2_inode_fifo))
+			goto bad_dinode;
 		break;
 	case S_IFSOCK:
 		if (fsck_blockmap_set(ip, block, _("socket"),
-				      gfs2_inode_sock)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
+				      gfs2_inode_sock))
+			goto bad_dinode;
 		break;
 	default:
 		/* We found a dinode that has an invalid mode, so we can't
@@ -997,19 +1046,12 @@ static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
 		   skip parts that we can't be sure of based on dinode type. */
 		check_metatree(ip, &invalidate_fxns);
 		if (fsck_blockmap_set(ip, block, _("invalid mode"),
-				      gfs2_inode_invalid)) {
-			stack;
-			fsck_inode_put(&ip);
-			return -1;
-		}
-		fsck_inode_put(&ip);
+				      gfs2_inode_invalid))
+			goto bad_dinode;
 		return 0;
 	}
-	if(set_link_count(ip->i_di.di_num.no_addr, ip->i_di.di_nlink)) {
-		stack;
-		fsck_inode_put(&ip);
-		return -1;
-	}
+	if(set_link_count(ip->i_di.di_num.no_addr, ip->i_di.di_nlink))
+		goto bad_dinode;
 
 	if (S_ISDIR(ip->i_di.di_mode) &&
 	    (ip->i_di.di_flags & GFS2_DIF_EXHASH)) {
@@ -1021,22 +1063,16 @@ static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
 				 ip->i_di.di_depth,
 				 (1 >> (ip->i_di.di_size/sizeof(uint64_t))));
 			if(fsck_blockmap_set(ip, block, _("bad depth"),
-					     gfs2_block_free)) {
-				stack;
-				fsck_inode_put(&ip);
-				return -1;
-			}
-			fsck_inode_put(&ip);
+					     gfs2_block_free))
+				goto bad_dinode;
 			return 0;
 		}
 	}
 
 	pass1_fxns.private = &bc;
 	error = check_metatree(ip, &pass1_fxns);
-	if (fsck_abort || error < 0) {
-		fsck_inode_put(&ip);
+	if (fsck_abort || error < 0)
 		return 0;
-	}
 	if (error > 0) {
 		log_err( _("Error: inode %llu (0x%llx) has unrecoverable "
 			   "errors; invalidating.\n"),
@@ -1050,7 +1086,6 @@ static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
 		   Therefore we mark the inode as free space. */
 		fsck_blockmap_set(ip, ip->i_di.di_num.no_addr,
 				  _("corrupt"), gfs2_block_free);
-		fsck_inode_put(&ip);
 		return 0;
 	}
 
@@ -1089,7 +1124,223 @@ static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
 				(unsigned long long)ip->i_di.di_num.no_addr);
 	}
 
+	return 0;
+bad_dinode:
+	stack;
+	return -1;
+}
+
+/*
+ * handle_di - This is now a wrapper function that takes a gfs2_buffer_head
+ *             and calls handle_ip, which takes an in-code dinode structure.
+ */
+static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh)
+{
+	uint8_t q;
+	int error = 0;
+	uint64_t block = bh->b_blocknr;
+	struct gfs2_inode *ip;
+
+	ip = fsck_inode_get(sdp, bh);
+	q = block_type(block);
+	if(q != gfs2_block_free) {
+		log_err( _("Found a duplicate inode block at #%" PRIu64
+			   " (0x%" PRIx64 ") previously marked as a %s\n"),
+			 block, block, block_type_string(q));
+		add_duplicate_ref(ip, block, ref_as_meta, 0, INODE_VALID);
+		fsck_inode_put(&ip);
+		return 0;
+	}
+
+	if (ip->i_di.di_num.no_addr != block) {
+		log_err( _("Inode #%llu (0x%llx): Bad inode address found: %llu "
+			"(0x%llx)\n"), (unsigned long long)block,
+			(unsigned long long)block,
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)ip->i_di.di_num.no_addr);
+		if(query( _("Fix address in inode at block #%"
+			    PRIu64 " (0x%" PRIx64 ")? (y/n) "),
+			  block, block)) {
+			ip->i_di.di_num.no_addr = ip->i_di.di_num.no_formal_ino = block;
+			bmodified(ip->i_bh);
+		} else
+			log_err( _("Address in inode at block #%" PRIu64
+				 " (0x%" PRIx64 ") not fixed\n"), block, block);
+	}
+	error = handle_ip(sdp, ip);
 	fsck_inode_put(&ip);
+	return error;
+}
+
+/* Check system inode and verify it's marked "in use" in the bitmap:       */
+/* Should work for all system inodes: root, master, jindex, per_node, etc. */
+/* We have to pass the sysinode as ** because the pointer may change out from
+   under the reference by way of the builder() function.  */
+static int check_system_inode(struct gfs2_sbd *sdp,
+			      struct gfs2_inode **sysinode,
+			      const char *filename,
+			      int builder(struct gfs2_sbd *sbp),
+			      enum gfs2_mark_block mark)
+{
+	uint64_t iblock = 0;
+	struct dir_status ds = {0};
+	int error;
+
+	log_info( _("Checking system inode '%s'\n"), filename);
+	if (*sysinode) {
+		/* Read in the system inode, look at its dentries, and start
+		 * reading through them */
+		iblock = (*sysinode)->i_di.di_num.no_addr;
+		log_info( _("System inode for '%s' is located at block %"
+			 PRIu64 " (0x%" PRIx64 ")\n"), filename,
+			 iblock, iblock);
+		if (gfs2_check_meta((*sysinode)->i_bh, GFS2_METATYPE_DI)) {
+			log_err( _("Found invalid system dinode at block #"
+				   "%llu (0x%llx)\n"),
+				 (unsigned long long)iblock,
+				 (unsigned long long)iblock);
+			gfs2_blockmap_set(bl, iblock, gfs2_block_free);
+			check_n_fix_bitmap(sdp, iblock, gfs2_block_free);
+			inode_put(sysinode);
+		}
+	}
+	if (*sysinode) {
+		ds.q = block_type(iblock);
+		/* If the inode exists but the block is marked free, we might
+		   be recovering from a corrupt bitmap.  In that case, don't
+		   rebuild the inode.  Just reuse the inode and fix the
+		   bitmap. */
+		if (ds.q == gfs2_block_free) {
+			log_info( _("The inode exists but the block is not "
+				    "marked 'in use'; fixing it.\n"));
+			fsck_blockmap_set(*sysinode,
+					  (*sysinode)->i_di.di_num.no_addr,
+					  filename, mark);
+			ds.q = mark;
+			if (mark == gfs2_inode_dir)
+				dirtree_insert((*sysinode)->i_di.di_num.no_addr);
+		}
+	} else
+		log_info( _("System inode for '%s' is corrupt or missing.\n"),
+			  filename);
+	/* If there are errors with the inode here, we need to create a new
+	   inode and get it all setup - of course, everything will be in
+	   lost+found then, but we *need* our system inodes before we can
+	   do any of that. */
+	if(!(*sysinode) || ds.q != mark) {
+		log_err( _("Invalid or missing %s system inode (should be %d, "
+			   "is %d).\n"), filename, mark, ds.q);
+		if (query(_("Create new %s system inode? (y/n) "), filename)) {
+			log_err( _("Rebuilding system file \"%s\"\n"),
+				 filename);
+			error = builder(sdp);
+			if (error) {
+				log_err( _("Error trying to rebuild system "
+					   "file %s: Cannot continue\n"),
+					 filename);
+				return error;
+			}
+			fsck_blockmap_set(*sysinode,
+					  (*sysinode)->i_di.di_num.no_addr,
+					  filename, mark);
+			ds.q = mark;
+			if (mark == gfs2_inode_dir)
+				dirtree_insert((*sysinode)->i_di.di_num.no_addr);
+		} else {
+			log_err( _("Cannot continue without valid %s inode\n"),
+				filename);
+			return -1;
+		}
+	}
+	if (S_ISDIR((*sysinode)->i_di.di_mode)) {
+		struct block_count bc = {0};
+
+		sysdir_fxns.private = &bc;
+		if ((*sysinode)->i_di.di_flags & GFS2_DIF_EXHASH)
+			check_metatree(*sysinode, &sysdir_fxns);
+		else
+			check_linear_dir(*sysinode, (*sysinode)->i_bh,
+					 &sysdir_fxns);
+	}
+	error = handle_ip(sdp, *sysinode);
+	return error;
+}
+
+static int build_a_journal(struct gfs2_sbd *sdp)
+{
+	char name[256];
+
+	/* First, try to delete the journal if it's in jindex */
+	sprintf(name, "journal%u", sdp->md.journals);
+	gfs2_dirent_del(sdp->md.jiinode, name, strlen(name));
+	/* Now rebuild it */
+	build_journal(sdp, sdp->md.journals, sdp->md.jiinode);
+	return 0;
+}
+
+static int check_system_inodes(struct gfs2_sbd *sdp)
+{
+	int journal_count;
+
+	/*******************************************************************
+	 *******  Check the system inode integrity             *************
+	 *******************************************************************/
+	if (check_system_inode(sdp, &sdp->master_dir, "master", build_master,
+			       gfs2_inode_dir)) {
+		stack;
+		return -1;
+	}
+	if (check_system_inode(sdp, &sdp->md.rooti, "root", build_root,
+			       gfs2_inode_dir)) {
+		stack;
+		return -1;
+	}
+	if (check_system_inode(sdp, &sdp->md.inum, "inum", build_inum,
+			       gfs2_inode_file)) {
+		stack;
+		return -1;
+	}
+	if (check_system_inode(sdp, &sdp->md.statfs, "statfs", build_statfs,
+			       gfs2_inode_file)) {
+		stack;
+		return -1;
+	}
+	if (check_system_inode(sdp, &sdp->md.jiinode, "jindex", build_jindex,
+			       gfs2_inode_dir)) {
+		stack;
+		return -1;
+	}
+	if (check_system_inode(sdp, &sdp->md.riinode, "rindex", build_rindex,
+			       gfs2_inode_file)) {
+		stack;
+		return -1;
+	}
+	if (check_system_inode(sdp, &sdp->md.qinode, "quota", build_quota,
+			       gfs2_inode_file)) {
+		stack;
+		return -1;
+	}
+	if (check_system_inode(sdp, &sdp->md.pinode, "per_node",
+			       build_per_node, gfs2_inode_dir)) {
+		stack;
+		return -1;
+	}
+	/* We have to play a trick on build_journal:  We swap md.journals
+	   in order to keep a count of which journal we need to build. */
+	journal_count = sdp->md.journals;
+	for (sdp->md.journals = 0; sdp->md.journals < journal_count;
+	     sdp->md.journals++) {
+		char jname[16];
+
+		sprintf(jname, "journal%d", sdp->md.journals);
+		if (check_system_inode(sdp, &sdp->md.journal[sdp->md.journals],
+				       jname, build_a_journal,
+				       gfs2_inode_file)) {
+			stack;
+			return -1;
+		}
+	}
+
 	return 0;
 }
 
@@ -1124,6 +1375,9 @@ int pass1(struct gfs2_sbd *sbp)
 	 * sweep - is there any metadata we need to mark here before
 	 * the sweeps start that we won't find otherwise? */
 
+	/* Make sure the system inodes are okay & represented in the bitmap. */
+	check_system_inodes(sbp);
+
 	/* So, do we do a depth first search starting at the root
 	 * inode, or use the rg bitmaps, or just read every fs block
 	 * to find the inodes?  If we use the depth first search, why
@@ -1172,6 +1426,14 @@ int pass1(struct gfs2_sbd *sbp)
 				skip_this_pass = FALSE;
 				fflush(stdout);
 			}
+			if (fsck_system_inode(sbp, block)) {
+				log_debug(_("Already processed system inode "
+					    "%lld (0x%llx)\n"),
+					  (unsigned long long)block,
+					  (unsigned long long)block);
+				first = 0;
+				continue;
+			}
 			bh = bread(sbp, block);
 
 			/*log_debug( _("Checking metadata block #%" PRIu64
