@@ -749,7 +749,6 @@ svc_start(const char *svcName, int req)
 	svcStatus.rs_transition = (uint64_t)time(NULL);
 
 	if (svcStatus.rs_state == RG_STATE_RECOVER) {
-		add_restart(svcName);
 		svcStatus.rs_restarts++;
 	} else {
 		svcStatus.rs_restarts = 0;
@@ -830,6 +829,11 @@ svc_migrate(const char *svcName, int target)
 		return RG_ENODE;
 	}
 
+	if (node_should_start_safe(target, membership, svcName) == FOD_ILLEGAL) {
+		free_member_list(membership);
+		return RG_EDEPEND;
+	}
+
 	count_resource_groups_local(m);
 	if (m->cn_svcexcl ||
 	    (m->cn_svccount && is_exclusive(svcName))) {
@@ -1745,6 +1749,7 @@ handle_relocate_req(char *svcName, int orig_request, int preferred_target,
 			 */
 			if (svc_start(svcName, RG_START) == 0) {
 				*new_owner = me;
+				free_member_list(backup);
 				return 0;
 			}
 		} else if (target == preferred_target) {
@@ -1757,6 +1762,7 @@ handle_relocate_req(char *svcName, int orig_request, int preferred_target,
 				/*
 				 * Great! We're done...
 				 */
+				free_member_list(backup);
 				return 0;
 			}
 
@@ -2082,20 +2088,25 @@ handle_recover_req(char *svcName, int *new_owner)
 	get_recovery_policy(svcName, policy, sizeof(policy));
 
 	if (!strcasecmp(policy, "disable")) {
+		clear_restart(svcName);
 		return svc_disable(svcName);
 	} else if (!strcasecmp(policy, "relocate")) {
+		clear_restart(svcName);
 		return handle_relocate_req(svcName, RG_START_RECOVER, -1,
 					   new_owner);
 	}
 
 	/* Check restart counter/timer for this resource */
 	if (check_restart(svcName) > 0) {
+		clear_restart(svcName);
 		logt_print(LOG_NOTICE, "Restart threshold for %s exceeded; "
 		       "attempting to relocate\n", svcName);
 		return handle_relocate_req(svcName, RG_START_RECOVER, -1,
 					   new_owner);
 	}
 
+	add_restart(svcName);
+
 	return handle_start_req(svcName, RG_START_RECOVER, new_owner);
 }
 
